#include <types/entity/entity.h>
#include <engine/occlusion.h>
#include <types/entity/camera.h>
#include <uuid_v4.h>
#include <engine/engine.h>

//inline LinearAlgebra::Vector3 Entity::GetPos() const
//{ return LinearAlgebra::Vector3(coordinates[1]); }

//void Entity::SetPos(const LinearAlgebra::Vector3& rhs)
//{ coordinates[1] = LinearAlgebra::Vector4(rhs.x, rhs.y, rhs.z, 0); }

//LinearAlgebra::Matrix4x4 Entity::GetMatrix() const
//{ return coordinates;}

void Entity::SetMatrix(const Matrix4x4& rhs)
{ coordinates = rhs; }

bool Entity::occluded() {
    Camera* camera;
    for (auto& e : engine->world->GetChildren())
        if (auto* c = dynamic_cast<Camera*>(e))
            camera = c;

    if (!Occlusion::frustumCull(camera, this))
    return false;
return true;
}

Vector3 Entity::GetPos() const {
    return position;
}

void Entity::SetPos(const Vector3 &rhs) {
    position = rhs;
}

Entity::Entity() {
    position = {0,0,0};
    ticksAlive = 0;
    children = std::vector<Entity*> ();
    parent = nullptr;
    scale = 1.0f;

    //TODO
    UUIDv4::UUIDGenerator<std::mt19937_64> uuidGenerator;
    UUIDv4::UUID id = uuidGenerator.getUUID();
    uuid = id.bytes();
}

void Entity::setScale(const float &multiplier) {
    scale = multiplier;
}

GLfloat Entity::getScale() const {
    return scale;
}

Collider Entity::getCollider() {
    //Default Collider is AABB.
    Collider collider = Collision::calculateAxisAlignedBoundingBox(getGeometry()->getBoundingVolume(angle));
    collider.angle = angle;
    //Scale.
    for (auto& vertex : collider.vertices) {
        vertex.x *= getScale();
        vertex.y *= getScale();
        vertex.z *= getScale();
    }

    //To world space.
    for (auto& vertex : collider.vertices) {
        vertex.x = vertex.x + position.x;
        vertex.y = vertex.y + position.y;
        vertex.z = vertex.z + position.z;
    }
    return collider;
}

void Entity::loadTexture() {
    Texture texture(name, texturePath.c_str(), true);
}

void Entity::loadGeometry() {
    VertexArray geometry;
    geometry.name = name;
    geometry.load(modelPath);
    engine->world->geometryList.push_back(geometry);
}

VertexArray* Entity::getGeometry() {
    if (engine->world->geometryList.empty())
        loadGeometry();
    for (auto& vArray : engine->world->geometryList)
        if (vArray.name == name) {
            return &vArray;
        }
    loadGeometry();
    return getGeometry();
}

//Default rendering routine.
void Entity::render() {
    bool cameraInside = false;
    bool multiTexture = false;

    //Camera inside the entity check.
    Camera* camera = nullptr;
    for (auto& e : engine->world->GetChildren())
        if (auto* c = dynamic_cast<Camera*>(e))
            camera = c;

    if (camera) {
        switch (getCollider().type) {
            case ColliderType::AxisAlignedBoundingBox:
                if (Collision::vertexInsideAxisAlignedBoundingBox(camera->position, getCollider()))
                    cameraInside = true;
                break;
            case ColliderType::OrientedBoundingBox:
                if (Collision::vertexInsideOrientedBoundingBox(camera->position, getCollider()))
                    cameraInside = true;
                break;
        }
    }
    if (cameraInside)
        glCullFace(GL_FRONT);

    //Done to avoid using the dynamic cast more than once because it's comparatively slow.
    if (auto* t = dynamic_cast<MultiTexture*>(this->getTexture()))
        multiTexture = true;

    if (engine->world->getGlobalFogMode() == NULL || getGeometry()->name == "skybox")
        glDisable(GL_FOG);

    glPushMatrix();
    glTranslatef(position.x, position.y, position.z);
    glRotatef(angle.x, 1.0f, 0.0f, 0.0f);
    glRotatef(angle.y, 0.0f, 1.0f, 0.0f);
    glRotatef(angle.z, 0.0f, 0.0f, 1.0f);
    if (getScale() != 1.0f)
        glScalef(getScale(), getScale(), getScale());

    //Set up primary texture.
    glActiveTexture(GL_TEXTURE0);
    glClientActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindTexture(GL_TEXTURE_2D, getTexture()->id);

    if (!engine->useVBO) {
        glVertexPointer(3, GL_FLOAT, sizeof(Vertex), &getGeometry()->vertices[0].x);
        glTexCoordPointer(2, GL_FLOAT, sizeof(TextureCoordinate), &getGeometry()->texCoords[0].x);
    }

    if (engine->useVBO) {
        glBindBuffer(GL_ARRAY_BUFFER, getGeometry()->tbo);
        glTexCoordPointer(2, GL_FLOAT, 0, nullptr);
    }

    if (!multiTexture) {
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
        getGeometry()->draw();
    }

    if (multiTexture) {
        auto* multi = (MultiTexture*) getTexture();

        int i = 0;
        for(auto& texture : multi->multi) {
            glActiveTexture(GL_TEXTURE1 + i);
            glClientActiveTexture(GL_TEXTURE1 + i);
            glEnable(GL_TEXTURE_2D);
            glEnableClientState(GL_VERTEX_ARRAY);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);

            if (!engine->useVBO) {
                glVertexPointer(3, GL_FLOAT, sizeof(Vertex), &getGeometry()->vertices[0].x);
                glTexCoordPointer(2, GL_FLOAT, sizeof(TextureCoordinate), &getGeometry()->texCoords[0].x);
            }

            if (engine->useVBO) {
                glBindBuffer(GL_ARRAY_BUFFER, getGeometry()->tbo);
                glTexCoordPointer(2, GL_FLOAT, 0, nullptr);
            }

            //Texture unit mode.
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);

            //Makes the lighting work the way you'd think it would.
            GLfloat blendingFactor[3] = {engine->world->getAmbientLightColor().x * 0.2f, engine->world->getAmbientLightColor().y * 0.2f, engine->world->getAmbientLightColor().z * 0.2f};
            glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, blendingFactor);
            glBindTexture(GL_TEXTURE_2D, texture.id);
            i++;
        }

        glEnable(GL_BLEND);
            getGeometry()->draw();
        glDisable(GL_BLEND);

        for (int j = 0; j <= i; j++) {
            glActiveTexture(GL_TEXTURE1 + j);
            glClientActiveTexture(GL_TEXTURE1 + j);
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
            glDisable(GL_TEXTURE_2D);
            glDisableClientState(GL_VERTEX_ARRAY);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }
        if (cameraInside)
            glCullFace(GL_BACK);
    }
    glActiveTexture(GL_TEXTURE0);
    glClientActiveTexture(GL_TEXTURE0);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glDisable(GL_TEXTURE_2D);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glPopMatrix();

    if (engine->debug) {
        glDisable(GL_LIGHTING);
        glPushMatrix();
        glColor4f(1, 0, 0, 1);
        getCollider().draw();
        glColor4f(1, 1, 1, 1);
        glPopMatrix();
        glEnable(GL_LIGHTING);
    }
}

bool Entity::isCollidingWith(Entity *entity) {
    Collider thisCollider = this->getCollider();
    Collider otherCollider = entity->getCollider();

    if (thisCollider.type == ColliderType::OrientedBoundingBox && otherCollider.type == ColliderType::OrientedBoundingBox)
        return Collision::orientedBoundingBoxCollides(thisCollider, otherCollider);

    if (thisCollider.type == ColliderType::AxisAlignedBoundingBox && otherCollider.type == ColliderType::AxisAlignedBoundingBox)
        return Collision::axisAlignedBoundingBoxCollides(thisCollider, otherCollider);

    if ((thisCollider.type == ColliderType::AxisAlignedBoundingBox && otherCollider.type == ColliderType::OrientedBoundingBox) || (thisCollider.type == ColliderType::OrientedBoundingBox && otherCollider.type == ColliderType::AxisAlignedBoundingBox))
        return Collision::orientedBoundingBoxCollides(thisCollider, otherCollider);

    if (thisCollider.type == ColliderType::AxisAlignedBoundingBox && otherCollider.type == ColliderType::CollisionMap)
        return Collision::collisionMapVSAxisAlignedBoundingBoxCollides(thisCollider, otherCollider);
}

Texture* Entity::getTexture() {
    for (auto& texture : engine->world->textureList)
        if (texture->name == name) {
                if (std::find(texture->usedBy.begin(), texture->usedBy.end(), this) == texture->usedBy.end())
                    texture->usedBy.push_back(this);
            return texture;
        }
    loadTexture();
    return getTexture();
}

void Entity::erase() {
    //Remove the usage reference from the list.
    Texture* t = getTexture();

    for (int i = 0; i < t->usedBy.size(); i++) {
        if (t->usedBy[i] == this) {
            t->usedBy.erase(t->usedBy.begin() + i);

            //If we just removed the last reference then the texture can be unloaded.
            if (t->usedBy.empty())
                t->erase();
        }
    }

    //TODO: ^^ Same thing for geometry.
    for (int i = 0; i < engine->world->GetChildren().size(); i++)
        if (engine->world->GetChildren()[i] == this)
            engine->world->GetChildren().erase(engine->world->GetChildren().begin() + i);
    delete this;
}
