#include <types/texture.h>
#include <SOIL/SOIL.h>
#include <engine/engine.h>

void Texture::load(const char* file, bool storeOnTextureList) {
    id = SOIL_load_OGL_texture(file, SOIL_LOAD_RGBA, SOIL_CREATE_NEW_ID, SOIL_FLAG_INVERT_Y);
    //If we can't load the missing texture
    if (file == "assets/textures/missing.png" && id == 0) {
        engine->setError(ENGINE_ERROR_CODE::TEXTURE_NOT_FOUND, true);
    }
    if (id == 0) {
        engine->setError(ENGINE_ERROR_CODE::TEXTURE_NOT_FOUND, false);
        load("assets/textures/missing.png", false);
    }
    if (storeOnTextureList)
        engine->world->textureList.push_back(new Texture(*this));
}

void Texture::erase() {
    Texture::erase(this);
}

Texture::Texture(std::string& name, const char *filePath, bool storeOnTextureList) {
    this->name = name;
    load(filePath, storeOnTextureList);
}

void Texture::erase(Texture* texture) const {
    glDeleteTextures(1, &texture->id);

    //If texture is being deleted while in use.
    //It won't crash the program. But it'll cause the texture to be reloaded from disk unnecessarily.
    if (!usedBy.empty())
        engine->setError(ENGINE_ERROR_CODE::TEXTURE_ERASED_WHILE_IN_USE, false);

    if (auto* t = dynamic_cast<MultiTexture*>(texture))
        for (auto &m : t->multi)
            glDeleteTextures(1, &m.id);

    for (int i = 0; i < engine->world->textureList.size(); i++)
        if (engine->world->textureList[i] == texture)
            engine->world->textureList.erase(engine->world->textureList.begin() + i);
}


MultiTexture::MultiTexture(const std::string& name, const char* pathContainingTextures, bool storeOnTextureList) {
    for (const auto& entry : std::filesystem::directory_iterator(pathContainingTextures))
        if (entry.is_regular_file() && entry.path().extension() == ".png") {
            Texture texture;
            texture.load(entry.path().c_str(), false);
            texture.name = name;

            //The base texture *must* go in the first slot.
            if (entry.path().filename() == "default.png") {
                this->id = texture.id;
                this->name = texture.name;
            } else {multi.push_back(texture);}
        }

    if (storeOnTextureList)
        engine->world->textureList.push_back(new MultiTexture(*this));

    //You cannot have more than 8 total textures rendered in one pass.
    //In Fixed-Function OpenGL you only have 8 TMUs available.
    //TODO: multi-pass multi-texturing (will be slower).
    if (multi.size() > 7)
        engine->setError(ENGINE_ERROR_CODE::MULTI_TEXTURE_SIZE_EXCEEDS, true);
}

void MultiTexture::erase() {
    Texture::erase(this);
}